package dev.shadowsoffire.apotheosis.adventure.socket.gem.bonus;

import java.util.function.Function;
import com.mojang.datafixers.kinds.Applicative;
import dev.shadowsoffire.placebo.codec.PlaceboCodecs;
import net.minecraftforge.registries.ForgeRegistries;
import com.mojang.datafixers.kinds.App;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import com.google.common.base.Preconditions;
import net.minecraft.world.effect.MobEffectUtil;
import dev.shadowsoffire.apotheosis.adventure.affix.Affix;
import net.minecraft.world.damagesource.DamageSource;
import net.minecraft.world.phys.EntityHitResult;
import net.minecraft.world.phys.HitResult;
import net.minecraft.world.entity.projectile.AbstractArrow;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraft.core.BlockPos;
import net.minecraft.world.level.LevelAccessor;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.entity.LivingEntity;
import net.minecraft.network.chat.MutableComponent;
import net.minecraft.world.effect.MobEffectInstance;
import net.minecraft.util.StringUtil;
import net.minecraft.ChatFormatting;
import net.minecraft.network.chat.Component;
import net.minecraft.world.item.ItemStack;
import dev.shadowsoffire.apotheosis.Apotheosis;
import dev.shadowsoffire.apotheosis.adventure.socket.gem.GemClass;
import dev.shadowsoffire.apotheosis.adventure.loot.LootRarity;
import java.util.Map;
import dev.shadowsoffire.apotheosis.adventure.affix.effect.PotionAffix;
import net.minecraft.world.effect.MobEffect;
import com.mojang.serialization.Codec;

public class PotionBonus extends GemBonus
{
    public static final Codec<PotionBonus> CODEC;
    protected final MobEffect effect;
    protected final PotionAffix.Target target;
    protected final Map<LootRarity, EffectData> values;
    protected final boolean stackOnReapply;
    
    public PotionBonus(final GemClass gemClass, final MobEffect effect, final PotionAffix.Target target, final Map<LootRarity, EffectData> values, final boolean stackOnReapply) {
        super(Apotheosis.loc("mob_effect"), gemClass);
        this.effect = effect;
        this.target = target;
        this.values = values;
        this.stackOnReapply = stackOnReapply;
    }
    
    @Override
    public Component getSocketBonusTooltip(final ItemStack gem, final LootRarity rarity) {
        final MobEffectInstance inst = this.values.get(rarity).build(this.effect);
        MutableComponent comp = this.target.toComponent(new Object[] { toComponent(inst) }).m_130940_(ChatFormatting.YELLOW);
        final int cooldown = this.getCooldown(rarity);
        if (cooldown != 0) {
            final Component cd = (Component)Component.m_237110_("affix.apotheosis.cooldown", new Object[] { StringUtil.m_14404_(cooldown) });
            comp = comp.m_130946_(" ").m_7220_(cd);
        }
        if (this.stackOnReapply) {
            comp = comp.m_130946_(" ").m_7220_((Component)Component.m_237115_("affix.apotheosis.stacking"));
        }
        return (Component)comp;
    }
    
    @Override
    public int getNumberOfUUIDs() {
        return 0;
    }
    
    public Codec<? extends GemBonus> getCodec() {
        return PotionBonus.CODEC;
    }
    
    @Override
    public boolean supports(final LootRarity rarity) {
        return this.values.containsKey(rarity);
    }
    
    @Override
    public void doPostHurt(final ItemStack gem, final LootRarity rarity, final LivingEntity user, final Entity attacker) {
        if (this.target == PotionAffix.Target.HURT_SELF) {
            this.applyEffect(gem, user, rarity);
        }
        else if (this.target == PotionAffix.Target.HURT_ATTACKER && attacker instanceof LivingEntity) {
            final LivingEntity tLiving = (LivingEntity)attacker;
            this.applyEffect(gem, tLiving, rarity);
        }
    }
    
    @Override
    public void doPostAttack(final ItemStack gem, final LootRarity rarity, final LivingEntity user, final Entity target) {
        if (this.target == PotionAffix.Target.ATTACK_SELF) {
            this.applyEffect(gem, user, rarity);
        }
        else if (this.target == PotionAffix.Target.ATTACK_TARGET && target instanceof LivingEntity) {
            final LivingEntity tLiving = (LivingEntity)target;
            this.applyEffect(gem, tLiving, rarity);
        }
    }
    
    @Override
    public void onBlockBreak(final ItemStack gem, final LootRarity rarity, final Player player, final LevelAccessor world, final BlockPos pos, final BlockState state) {
        if (this.target == PotionAffix.Target.BREAK_SELF) {
            this.applyEffect(gem, (LivingEntity)player, rarity);
        }
    }
    
    @Override
    public void onArrowImpact(final ItemStack gemStack, final LootRarity rarity, final AbstractArrow arrow, final HitResult res) {
        if (this.target == PotionAffix.Target.ARROW_SELF) {
            final Entity 19749_ = arrow.m_19749_();
            if (19749_ instanceof final LivingEntity owner) {
                this.applyEffect(gemStack, owner, rarity);
            }
        }
        else if (this.target == PotionAffix.Target.ARROW_TARGET && res.m_6662_() == HitResult.Type.ENTITY) {
            final Entity 82443_ = ((EntityHitResult)res).m_82443_();
            if (82443_ instanceof final LivingEntity target) {
                this.applyEffect(gemStack, target, rarity);
            }
        }
    }
    
    @Override
    public float onShieldBlock(final ItemStack gem, final LootRarity rarity, final LivingEntity entity, final DamageSource source, final float amount) {
        if (this.target == PotionAffix.Target.BLOCK_SELF) {
            this.applyEffect(gem, entity, rarity);
        }
        else if (this.target == PotionAffix.Target.BLOCK_ATTACKER) {
            final Entity 7640_ = source.m_7640_();
            if (7640_ instanceof final LivingEntity target) {
                this.applyEffect(gem, target, rarity);
            }
        }
        return amount;
    }
    
    protected int getCooldown(final LootRarity rarity) {
        final EffectData data = this.values.get(rarity);
        return data.cooldown;
    }
    
    private void applyEffect(final ItemStack gemStack, final LivingEntity target, final LootRarity rarity) {
        final int cooldown = this.getCooldown(rarity);
        if (cooldown != 0 && Affix.isOnCooldown(this.getCooldownId(gemStack), cooldown, target)) {
            return;
        }
        final EffectData data = this.values.get(rarity);
        final MobEffectInstance inst = target.m_21124_(this.effect);
        if (this.stackOnReapply && inst != null) {
            if (inst != null) {
                final MobEffectInstance newInst = new MobEffectInstance(this.effect, Math.max(inst.m_19557_(), data.duration), inst.m_19564_() + 1 + data.amplifier);
                target.m_7292_(newInst);
            }
        }
        else {
            target.m_7292_(data.build(this.effect));
        }
        Affix.startCooldown(this.getCooldownId(gemStack), target);
    }
    
    public static Component toComponent(final MobEffectInstance inst) {
        MutableComponent mutablecomponent = Component.m_237115_(inst.m_19576_());
        final MobEffect mobeffect = inst.m_19544_();
        if (inst.m_19564_() > 0) {
            mutablecomponent = Component.m_237110_("potion.withAmplifier", new Object[] { mutablecomponent, Component.m_237115_("potion.potency." + inst.m_19564_()) });
        }
        if (inst.m_19557_() > 20) {
            mutablecomponent = Component.m_237110_("potion.withDuration", new Object[] { mutablecomponent, MobEffectUtil.m_267641_(inst, 1.0f) });
        }
        return (Component)mutablecomponent.m_130940_(mobeffect.m_19483_().m_19497_());
    }
    
    @Override
    public PotionBonus validate() {
        Preconditions.checkNotNull((Object)this.effect, (Object)"Null mob effect");
        Preconditions.checkNotNull((Object)this.target, (Object)"Null target");
        Preconditions.checkNotNull((Object)this.values, (Object)"Null values map");
        return this;
    }
    
    static {
        CODEC = RecordCodecBuilder.create(inst -> inst.group((App)GemBonus.gemClass(), (App)ForgeRegistries.MOB_EFFECTS.getCodec().fieldOf("mob_effect").forGetter(a -> a.effect), (App)PotionAffix.Target.CODEC.fieldOf("target").forGetter(a -> a.target), (App)LootRarity.mapCodec((Codec)EffectData.CODEC).fieldOf("values").forGetter(a -> a.values), (App)PlaceboCodecs.nullableField((Codec)Codec.BOOL, "stack_on_reapply", (Object)false).forGetter(a -> a.stackOnReapply)).apply((Applicative)inst, PotionBonus::new));
    }
    
    record EffectData(int duration, int amplifier, int cooldown) {
        private static Codec<EffectData> CODEC;
        
        public MobEffectInstance build(final MobEffect effect) {
            return new MobEffectInstance(effect, this.duration, this.amplifier);
        }
        
        static {
            EffectData.CODEC = (Codec<EffectData>)RecordCodecBuilder.create(inst -> inst.group((App)Codec.INT.fieldOf("duration").forGetter((Function)EffectData::duration), (App)Codec.INT.fieldOf("amplifier").forGetter((Function)EffectData::amplifier), (App)PlaceboCodecs.nullableField((Codec)Codec.INT, "cooldown", (Object)0).forGetter((Function)EffectData::cooldown)).apply((Applicative)inst, EffectData::new));
        }
    }
}
